use std::{
    error::Error,
    io,
    time::{Duration, Instant},
};

use clap::Parser;
use clock_tui::app::Mode;
use crossterm::{
    event::{self, Event, KeyCode},
    execute,
    terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
};
use tui::{
    backend::{Backend, CrosstermBackend},
    Terminal,
};

use clock_tui::app::App;

fn run_app<B: Backend>(
    terminal: &mut Terminal<B>,
    app: &mut App,
    tick_rate: Duration,
) -> io::Result<()> {
    let mut last_tick = Instant::now();

    loop {
        if app.is_ended() {
            return Ok(());
        }

        // draw the TUI using the `tui` library
        terminal.draw(|f| app.ui(f))?;

        let timeout = tick_rate
            .checked_sub(last_tick.elapsed()) // substraction, is None if would underflow (be negative)
            .unwrap_or(Duration::ZERO); // if it was `None`, it was longer than `tick_rate` since the
                                        // `last_tick`. We set the timeout to `DURATION::ZERO` so
                                        // that `event::poll` will return instantly with success

        // wait up to 100 ms for the timeout Duration. This ensures that we do not loop very fast,
        // which would cause high cpu load. If the Duration is less than 100 ms, we enter the if
        // clause. An event describes some kind of user interaction: Mouse actions, Key Presses, or
        // Resizing of the Terminal
        if event::poll(timeout)? {
            // we only care about pressed keys
            if let Event::Key(key) = event::read()? {
                match key.code {
                    // exit the application if the user presses the 'q' key
                    KeyCode::Char('q') => {
                        return Ok(());
                    }
                    // otherwise, let the app handle it the key that was pressed
                    key => app.on_key(key),
                }
            }
        }

        // has it been longer since the last tick than the duration of the `tick_rate`?
        if last_tick.elapsed() >= tick_rate {
            // if so, we want to substract with NOW when we calculate the new timeout.
            last_tick = Instant::now();
        }
    }
}

// entry point into the program
fn main() -> Result<(), Box<dyn Error>> {
    // Parse the cli arguments
    let mut app = App::parse();
    app.init_app();

    // setup terminal
    enable_raw_mode()?;
    let mut stdout = io::stdout();
    execute!(stdout, EnterAlternateScreen /*EnableMouseCapture*/,)?;
    let backend = CrosstermBackend::new(stdout);
    let mut terminal = Terminal::new(backend)?;

    // create app and run it
    let low_rate = match app.mode {
        Some(Mode::Clock {
            millis, no_seconds, ..
        }) => !millis || no_seconds,
        Some(Mode::Timer { no_millis, .. }) => no_millis,
        Some(Mode::Countdown { millis, .. }) => !millis,
        _ => false,
    };
    let tick_rate = Duration::from_millis(if low_rate { 200 } else { 20 });
    let res = run_app(&mut terminal, &mut app, tick_rate);

    // restore terminal
    disable_raw_mode()?;
    execute!(
        terminal.backend_mut(),
        LeaveAlternateScreen,
        // DisableMouseCapture
    )?;
    terminal.show_cursor()?;

    if let Err(err) = res {
        eprintln!("{:?}", err)
    }

    Ok(())
}
